# images.pl - build the eeprom images.
#
# this takes an input and an output file and processes all images in the
# input file into an EEPROM image in the specified output file.
#
# this goes along with the code in URL.ASM which assumes that EEPROM location
# 0000 contains the base of the image directory.
#
$debug = 0 ;
$MAX_BYTES = 8000 ;   # Max. payload of Ethernet GET response packet
$ver_file = '';
$version = '??';

$ego = $0;  $ego =~ s/.*[\\\/]//;
sub usage {
    local($why) = @_;

    print STDERR "usage: $ego [options] project_file images.dat userdef.h prefix\n";
    print STDERR "options:\n";
    print STDERR "  -d         - debug mode\n";
    print STDERR "  -v file    - get version info from 'file'\n";
    print STDERR "'project' is PicoWeb project definition file\n";
    print STDERR "'images.dat' output SEEPROM data file\n";
    print STDERR "'userdef.h' is output file for user defines\n";
    print STDERR "'prefix' is prefix for user-specific .asm files\n";
    print STDERR "$why\n" if ($why ne '');
    exit 1;
}

while ($ARGV[0] =~ /^-/) {
	$opt = $ARGV[0]; shift;
	if ($opt eq '-d') {
		$debug = 1;
	} elsif ($opt eq '-v') {
		$ver_file = $ARGV[0]; shift;
	} elsif ($opt eq '-h') {
		&usage('');
	} else {
		print STDERR "$ego: unknown option '$opt'!\n";
		exit 1;
	}
}

$proj_file = $ARGV[0]; shift; &usage('missing project file') if ($proj_file eq '');
$data_file = $ARGV[0]; shift; &usage('missing images.dat file') if ($data_file eq '');
$def_file = $ARGV[0]; shift; &usage('missing userdef.h file') if ($def_file eq '');
$prefix = $ARGV[0]; shift; &usage('missing prefix') if ($prefix eq '');

%avr_def2file = (
    '#avr_asm',   $prefix.'asm.asm',
    '#avr_reset', $prefix.'rst.asm',
    '#avr_slow',  $prefix.'sid.asm',
    '#avr_fast',  $prefix.'fid.asm'
);

%avr_def2def = (
    '#avr_asm',   'USER_ASM',
    '#avr_reset', 'USER_RESET_ASM',
    '#avr_slow',  'USER_SLOW_IDLE_ASM',
    '#avr_fast',  'USER_FAST_IDLE_ASM'
);

$hdr = "\n\n\n\n\n// DO NOT EDIT THIS FILE!!!!! (Auto generated by $ego!)\n\n\n\n\n";

open(PROJ,$proj_file) || die "$ego: Can't open project file '$proj_file'\n" ;
open(OUT,">$data_file") || die "$ego: Can't create output data file '$data_file'\n" ;
binmode(OUT) ;

open(DEFS,">$def_file") || die "$ego: Can't create output defines file '$def_file'\n" ;
print DEFS $hdr;

$imgdir = "" ;
$imgdata = "" ;
$numimages = 0 ;
$imgloc = 0 ;
$found_avr_asm = '';

if ($ver_file ne '') {
    if (open(V,"<$ver_file"))  {
        while (<V>) {
            $line = $_;
            if ($line =~ /^[#]define\s+VERSION\s+"([^"]+)"/) {
                $version =$1 ;
            }
        }
        close(V) ;
    } else {
        print "can't open version file '$ver_file'!\n";
    }
}
print "Version: $version\n";

$image_number = 0 ;
$index = 0 ;
$num_asms = 0;
$line_num = 0;
while (<PROJ>)  {
    ++$line_num;
    $line = $_;
    $line =~ s/[\s\r\n]+$//;
    $org_line = $line;
    $line =~ s/\/\/.*//;	# remove // comments
    $line =~ s/\s+$//;
    if ($line =~ /^([#]avr_[a-z]+)$/) {
        if ($avr_def2file{$1} ne '') {
            $found_avr_asm = $1;
            last;
        } else {
            print STDERR "$ego: Bad directive ($1)!\n";
            exit 1;
        }
    } elsif ($line =~ /^[#]/) {
        print DEFS "$org_line\n";
    } elsif ($line =~ /^([a-zA-Z0-9_]+)[.]([a-zA-Z0-9]+)$/) {
	$base = $1;
	$ext = $2;
	($file,$x) = split(/\s/, $line) ;
        $images{$image_number} = $file ;
        if ($ext =~ /CGI/i) {
            $indexh = sprintf("%02X",$index) ;
            $image_name_to_number{"\"$base.cgi\""} = "\"iu$indexh\"" ;
            $image_name_to_number{"\`$base.cgi\`"} = "\`7$indexh" ;
            $cgi_map[$index] = $base;
            $index++ ;
        } elsif ($ext =~ /ASM/i) {
            $asms[$num_asms++] = $line;
	} else {
            $base = $file; $base =~ s/.*[\\\/]//;
            $image_name_to_number{"\"$base\""} = sprintf("\"ii%02X\"",$image_number) ;
            $image_number++ ;
	} 
        print DEFS "// $org_line\n";
    } else {
        if ($org_line =~ /^\s*\/\//) {
            $s = "$org_line\n";
        } else {
            $s = "// $org_line\n";
        }
        print DEFS $s;
    }
}

&check_avr_def('#avr_asm');
print ASM $hdr;
print ASM ".eseg\n";
print ASM ".org 0x8000+CGI_INDEX_BASE\n\n";
print ASM "#define CGI(X) .dw X,(0x8000+0x4000)    ; flags as CGI\n";
print ASM "cgi_table:           // CGI code lookup table\n";
print ASM "    .dw (cgi_table_end-cgi_table)/4\n";
for ($i = 0; $i < $index; ++$i) {
    printf(ASM  "    CGI(%-32s    ; iu%02x\n", "$cgi_map[$i])", $i);
}
print ASM "cgi_table_end:\n";

if ($num_asms) {
    for ($i = 0; $i < $num_asms; ++$i) {
        print ASM "#include \"$asms[$i]\"\n";
    }
    print ASM "\n";
}
print ASM "#line $line_num \"$proj_file\"\n";

&check_avr_def($found_avr_asm);
while (<PROJ>)  {
    ++$line_num;
    $line = $_;
    $line =~ s/[\s\r\n]+$//;
    if ($line =~ /^([#]avr_[a-z]+)$/) {
        if ($avr_def2file{$1} ne '') {
            $found_avr_asm = $1;
            if ($last_avr_directive ne $found_avr_asm) {
                &check_avr_def($found_avr_asm);
            }
            print ASM "// $1 -> $avr_def2file{$1}\n";
            next;
        } else {
            print STDERR "$ego: Bad directive ($1)!\n";
            exit 1;
        }
    }
    print ASM "$line\n";
}
close(PROJ) ;

if ($debug)  {
    while (($k,$v) = each %image_name_to_number)  {
        print "$k\t$v\n" ;
    }
}

$number_of_images = $image_number ;
for ($image_number=0 ; $image_number < $number_of_images ; $image_number++)  {
    $file = $images{$image_number} ;
    if ($file =~ /\.jpg$/i)  {
        $flag = 0 ;
    }
    elsif ($file =~ /\.gif$/i)  {
        $flag = 0x8000 ;
    }
    elsif ($file =~ /\.htm$/i)  {
    	$flag = 0x4000 ;
    }
    elsif ($file =~ /\.html$/i)  {
    	$flag = 0x4000 ;
    }
    else  {
        print STDERR "$ego: unknown file type ($file) in PicoWeb project!\n" ;
        exit 1 ;              # not an image we know, blow it off
    }
    $pat = '$$VERSION$$';
    $name = $file ;
    open(F,"<$file") || die "Can't open $file" ;
    binmode(F) ;
    sysread(F,$str,16384) ; # haul in the whole thing
    $len = length($str);
    printf STDOUT "ii%02X: $name ($len bytes)\n",$numimages ;
    if ($len > $MAX_BYTES) {
        print "Warning! Truncating '$file' from $len to $MAX_BYTES bytes!\n" ;
        $str = substr($str, 0, $MAX_BYTES) ;
    }
    if ($flag == 0x4000)  {
        $str =~ s/[\r]//g ;
        $str =~ s/[\n]/\r/g ;
        while ($str =~ s/\r\/\/[^\r]*\r/\r/g) { ; } # simple // comments
        ##$str =~ s/([^:])\/\/[^\r]*\r/$1\r/g ; # simple // comments
        ###$str =~ s/[$][$]VERSION[$][$]/$version/g ;
        $i = index($str, $pat);
        if ($i >= 0) {
            $str = substr($str,0,$i).$version.substr($str,$i+length($pat));
        }
        $str =~ s/(\"[^\"]+\")/&nametonum($1)/eg ;
        $str =~ s/(`[^`]+\.cgi`)/&nametonum($1)/eg ;
        while ($str =~ /^(.*[`])[?]([0-9a-f][[0-9a-f][^0-9a-f].*)$/i) {
           $str = $1 . '0' . $2;   # map `?hh --> `0hh
        }
        $str =~ s/[\r]/\n/g ;
	print "\n-------  $file  ------\n$str" if ($debug) ;
        $str .= pack("C",0) ;   # make sure HTML has a trailing NUL
    }
    $imgdata .= $str ;      # append the data
    $offsets{$numimages} = $imgloc ;
    $lengths{$numimages} = length($str) ;
    $flags{$numimages} = $flag ;
    $numimages++ ;
    $imgloc += length($str) ;
    close(F) ;
}
$imgdir = pack("S",$numimages) ;    # # of images
$imgbase = 2+($numimages*4) ;       # first location
print "$numimages images\n" if ($debug);
print "$imgloc bytes of HTML code/images total\n";

for ($i=0 ; $i<$numimages ; $i++)  {
    $imgdir .= pack("SS",$offsets{$i}+$imgbase,$lengths{$i}|$flags{$i}) ;
    if ($debug) {
        print "$i:\t$offsets{$i}+$imgbase, $lengths{$i} | $flags{$i}\n";
        printf("\t%04x, %04x | %04x\n",
            $offsets{$i}+$imgbase, $lengths{$i}, $flags{$i});
    }
}
$imgdir .= $imgdata ;
if (syswrite(OUT,$imgdir,length($imgdir)) != length($imgdir))  {
    die "Can't write output length ",length($imgdir),"\n" ;
}
close(OUT) ;
close(DEFS) ;
close(ASM) ;
exit 0;

sub check_avr_def {
    local($avr_directive) = @_ ;
    local($s, $def) ;

    return if ($avr_directive eq $last_avr_directive);
    $def = $avr_def2def{$avr_directive};
    $asm_file = $avr_def2file{$avr_directive};
    if ($asm_file eq '') {
        die "$ego: internal error!  Cannot find file for $avr_directive";
    }
    if (!$avr_defs{$def}) {
        print DEFS "#define $def \"$asm_file\"\n";      
        ++$avr_defs{$def};
    }
    if ($last_avr_directive ne '') {
        $s = '>';
        close(ASM)
    }
    open(ASM,"$s>$asm_file") || die "$ego: Can't create output ASM file '$asm_file'\n" ;
    if ($last_avr_directive ne '') {
        print ASM "#line $line_num \"$proj_file\"\n";
    }
    $last_avr_directive = $avr_directive;
}

sub nametonum  {
    local($name) = @_ ;

    if (defined $image_name_to_number{$name})  {
        print "Mapped $name to ",$image_name_to_number{$name},"\n" if ($debug);
        return $image_name_to_number{$name} ;
    }
    else  {
        print "skipping '$name'\n" if ($debug) ;
        return $name ;
    }
}
